#ifndef _Inheritance_and_genericity_h
#define _Inheritance_and_genericity_h

#include <bitset>
#include <type_traits>
#include <unordered_set>

template<typename Any> class Pointer_based_hash {
public:
    // If 'std::is_pointer<Any>::value' is true then 'std::enable_if' has a public member 'typedef' type, which equals 'Return_type' else there is no member 'typedef' 

    template<typename Return_type = std::size_t> typename std::enable_if<std::is_pointer<Any>::value, Return_type>::type
    operator()(const Any pointer) const {
        std::hash < std::bitset<sizeof (Any)> > hash;
        return hash(reinterpret_cast<uintptr_t> (pointer)); // Storage relies on memory address!
    }

    template<typename Return_type = std::size_t> typename std::enable_if<!std::is_pointer<Any>::value, Return_type>::type
    operator()(const Any& any) const {
        std::hash < std::bitset<sizeof (Any*)> > hash;
        return hash(reinterpret_cast<uintptr_t> (&any)); // Storage relies on memory address!
    }
};

template <typename T> class Garden {
private:
    // 'Pointer_based_hash<T>' (as second anonymous type) acts as a functor (definition of 'operator()' required in 'Pointer_based_hash<T>'):
    std::unordered_set <T, Pointer_based_hash<T> > _garden;
public:
    Garden(const T&);
};

template <typename T> Garden<T>::Garden(const T& t) {
    _garden.insert(t);
}

class Vegetable {
public:

    //    std::size_t operator()(const Vegetable& vegetable) const {
    //        // Hash on several fields uses '^': https://www.techiedelight.com/use-pair-key-std-unordered_set-cpp/
    //        std::hash < std::bitset<sizeof (Vegetable*)> > hash;
    //        return hash(reinterpret_cast<uintptr_t> (&vegetable)); // Storage relies on memory address!
    //    }

    enum class Vegetable_colors {
        Green, Orange, Purple, Red
    };
    virtual Vegetable_colors color() const;
};

class Cabbage : public Vegetable {
public:
    bool operator==(const Cabbage&) const;
    bool operator!=(const Cabbage&) const;
};

class Carot : public Vegetable {
public:
    bool operator==(const Carot&) const;
    bool operator!=(const Carot&) const;
    Vegetable_colors color() const;
};

#endif
