#include <iostream>

#include "Inheritance_and_genericity.h"

Vegetable::Vegetable_colors Vegetable::color() const {
    return Vegetable_colors::Green; // Default color...
}

bool Cabbage::operator==(const Cabbage& cabbage) const {
    return std::is_same<decltype(*this), decltype(cabbage)>::value; // For simplification
}

bool Cabbage::operator!=(const Cabbage& cabbage) const {
    return !this->operator==(cabbage);
}

bool Carot::operator==(const Carot& carot) const {
    return std::is_same<decltype(*this), decltype(carot)>::value; // For simplification
}

bool Carot::operator!=(const Carot& carot) const {
    return !this->operator==(carot);
}

Vegetable::Vegetable_colors Carot::color() const {
    return Vegetable_colors::Orange;
}

bool operator==(const Carot& carot, const Cabbage& cabbage) {
    return true; // "Mélanger des choux et des carottes !"
}

bool operator==(const Cabbage& cabbage, const Carot& carot) {
    return false; // "Mélanger des carottes et des choux !"
}

int main(int argc, char** argv) {
    Cabbage cabbage;

    Garden<Cabbage> garden(cabbage);
    // There is a risk to add carrots for instance, while 'another_garden' refers to a garden of cabbages only:
    // Garden<Vegetable>& another_garden = garden; // This *DOES NOT* work (contravariance)...

    Garden<Cabbage*> garden_(&cabbage);
    // There is a risk to add carrots for instance, while 'another_garden' refers to a garden of cabbages only:
    Garden<Vegetable*>& another_garden_ = garden_; // This *DOES NOT* work (contravariance)...

    // Generic and polymorphic lambda (C++14) -> https://stackoverflow.com/questions/26435052/using-template-parameter-in-a-generic-lambda
    // Visual Studio 2019 is not able to deal with 'decltype(a)':
    auto equals_generic = [](auto a, decltype(a) b) {
        return a == b;
    };
    auto equals_polymorphic = [](auto a, auto b) {
        return a == b;
    };

    std::cout << std::boolalpha << equals_generic(Cabbage{}, Cabbage{}) << std::endl;
    std::cout << std::boolalpha << equals_generic(Carot{}, Carot{}) << std::endl;
    // std::cout << std::boolalpha << equals_generic(Cabbage{}, Carot{}) << std::endl; // Compilation error due to non-polymorphic
    // std::cout << std::boolalpha << equals_generic(Carot{}, Cabbage{}) << std::endl; // Compilation error due to non-polymorphic:

    std::cout << std::boolalpha << equals_polymorphic(Carot{}, Carot{}) << std::endl;
    std::cout << std::boolalpha << equals_polymorphic(Cabbage{}, Cabbage{}) << std::endl;
    std::cout << std::boolalpha << equals_polymorphic(Cabbage{}, Carot{}) << std::endl;
    std::cout << std::boolalpha << equals_polymorphic(Carot{}, Cabbage{}) << std::endl;

    return 0;
}

